@@ -10,6 +10,11 @@ from lxml import etree
 __all__ = ['SelectorSyntaxError', 'ExpressionError',
            'CSSSelector']
 
+try:
+    _basestring = basestring
+except NameError:
+    _basestring = str
+
 class SelectorSyntaxError(SyntaxError):
     pass
 
@@ -27,10 +32,24 @@ class CSSSelector(etree.XPath):
         >>> root = etree.XML("<a><b><c/><tag><child>TEXT</child></tag></b></a>")
         >>> [ el.tag for el in select(root) ]
         ['child']
+    
+    To use CSS namespaces, you need to pass a prefix-to-namespace
+    mapping as ``namespaces`` keyword argument::
+    
+        >>> rdfns = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
+        >>> select_ns = cssselect.CSSSelector('root > rdf|Description',
+        ...                                   namespaces={'rdf': rdfns})
+
+        >>> rdf = etree.XML((
+        ...     '<root xmlns:rdf="%s">'
+        ...       '<rdf:Description>blah</rdf:Description>'
+        ...     '</root>') % rdfns)
+        >>> [(el.tag, el.text) for el in select_ns(rdf)]
+        [('{http://www.w3.org/1999/02/22-rdf-syntax-ns#}Description', 'blah')]
     """
-    def __init__(self, css):
+    def __init__(self, css, namespaces=None):
         path = css_to_xpath(css)
-        etree.XPath.__init__(self, path)
+        etree.XPath.__init__(self, path, namespaces=namespaces)
         self.css = css
 
     def __repr__(self):
@@ -42,16 +61,24 @@ class CSSSelector(etree.XPath):
 ##############################
 ## Token objects:
 
-class _UniToken(unicode):
+try:
+    _unicode = unicode
+    _unichr = unichr
+except NameError:
+    # Python 3
+    _unicode = str
+    _unichr = chr
+
+class _UniToken(_unicode):
     def __new__(cls, contents, pos):
-        obj = unicode.__new__(cls, contents)
+        obj = _unicode.__new__(cls, contents)
         obj.pos = pos
         return obj
         
     def __repr__(self):
         return '%s(%s, %r)' % (
             self.__class__.__name__,
-            unicode.__repr__(self),
+            _unicode.__repr__(self),
             self.pos)
 
 class Symbol(_UniToken):
@@ -88,7 +115,7 @@ class Class(object):
     def xpath(self):
         sel_xpath = self.selector.xpath()
         sel_xpath.add_condition(
-            "contains(concat(' ', normalize-space(@class), ' '), %s)" % xpath_repr(' '+self.class_name+' '))
+            "contains(concat(' ', normalize-space(@class), ' '), %s)" % xpath_literal(' '+self.class_name+' '))
         return sel_xpath
 
 class Function(object):
@@ -115,11 +142,11 @@ class Function(object):
         sel_path = self.selector.xpath()
         if self.name in self.unsupported:
             raise ExpressionError(
-                "The psuedo-class %r is not supported" % self.name)
+                "The pseudo-class %r is not supported" % self.name)
         method = '_xpath_' + self.name.replace('-', '_')
         if not hasattr(self, method):
             raise ExpressionError(
-                "The psuedo-class %r is unknown" % self.name)
+                "The pseudo-class %r is unknown" % self.name)
         method = getattr(self, method)
         return method(sel_path, self.expr)
 
@@ -183,7 +210,7 @@ class Function(object):
         if isinstance(expr, Element):
             expr = expr._format_element()
         xpath.add_condition('contains(css:lower-case(string(.)), %s)'
-                            % xpath_repr(expr.lower()))
+                            % xpath_literal(expr.lower()))
         # FIXME: Currently case insensitive matching doesn't seem to be happening
         return xpath
 
@@ -227,11 +254,11 @@ class Pseudo(object):
         el_xpath = self.element.xpath()
         if self.ident in self.unsupported:
             raise ExpressionError(
-                "The psuedo-class %r is unsupported" % self.ident)
+                "The pseudo-class %r is unsupported" % self.ident)
         method = '_xpath_' + self.ident.replace('-', '_')
         if not hasattr(self, method):
             raise ExpressionError(
-                "The psuedo-class %r is unknown" % self.ident)
+                "The pseudo-class %r is unknown" % self.ident)
         method = getattr(self, method)
         el_xpath = method(el_xpath)
         return el_xpath
@@ -338,34 +365,34 @@ class Attrib(object):
             path.add_condition(attrib)
         elif self.operator == '=':
             path.add_condition('%s = %s' % (attrib,
-                                            xpath_repr(value)))
+                                            xpath_literal(value)))
         elif self.operator == '!=':
             # FIXME: this seems like a weird hack...
             if value:
                 path.add_condition('not(%s) or %s != %s'
-                                   % (attrib, attrib, xpath_repr(value)))
+                                   % (attrib, attrib, xpath_literal(value)))
             else:
                 path.add_condition('%s != %s'
-                                   % (attrib, xpath_repr(value)))
-            #path.add_condition('%s != %s' % (attrib, xpath_repr(value)))
+                                   % (attrib, xpath_literal(value)))
+            #path.add_condition('%s != %s' % (attrib, xpath_literal(value)))
         elif self.operator == '~=':
-            path.add_condition("contains(concat(' ', normalize-space(%s), ' '), %s)" % (attrib, xpath_repr(' '+value+' ')))
+            path.add_condition("contains(concat(' ', normalize-space(%s), ' '), %s)" % (attrib, xpath_literal(' '+value+' ')))
         elif self.operator == '|=':
             # Weird, but true...
             path.add_condition('%s = %s or starts-with(%s, %s)' % (
-                attrib, xpath_repr(value),
-                attrib, xpath_repr(value + '-')))
+                attrib, xpath_literal(value),
+                attrib, xpath_literal(value + '-')))
         elif self.operator == '^=':
             path.add_condition('starts-with(%s, %s)' % (
-                attrib, xpath_repr(value)))
+                attrib, xpath_literal(value)))
         elif self.operator == '$=':
             # Oddly there is a starts-with in XPath 1.0, but not ends-with
             path.add_condition('substring(%s, string-length(%s)-%s) = %s'
-                               % (attrib, attrib, len(value)-1, xpath_repr(value)))
+                               % (attrib, attrib, len(value)-1, xpath_literal(value)))
         elif self.operator == '*=':
             # FIXME: case sensitive?
             path.add_condition('contains(%s, %s)' % (
-                attrib, xpath_repr(value)))
+                attrib, xpath_literal(value)))
         else:
             assert 0, ("Unknown operator: %r" % self.operator)
         return path
@@ -414,7 +441,7 @@ class Hash(object):
 
     def xpath(self):
         path = self.selector.xpath()
-        path.add_condition('@id = %s' % xpath_repr(self.id))
+        path.add_condition('@id = %s' % xpath_literal(self.id))
         return path
 
 class Or(object):
@@ -467,7 +494,7 @@ class CombinedSelector(object):
 
     def _xpath_descendant(self, xpath, sub):
         # when sub is a descendant in any way of xpath
-        xpath.join('/descendant::', sub.xpath())
+        xpath.join('/descendant-or-self::*/', sub.xpath())
         return xpath
     
     def _xpath_child(self, xpath, sub):
@@ -490,12 +517,12 @@ class CombinedSelector(object):
 ##############################
 ## XPathExpr objects:
 
-_el_re = re.compile(r'^\w+\s*$')
-_id_re = re.compile(r'^(\w*)#(\w+)\s*$')
-_class_re = re.compile(r'^(\w*)\.(\w+)\s*$')
+_el_re = re.compile(r'^\w+\s*$', re.UNICODE)
+_id_re = re.compile(r'^(\w*)#(\w+)\s*$', re.UNICODE)
+_class_re = re.compile(r'^(\w*)\.(\w+)\s*$', re.UNICODE)
 
 def css_to_xpath(css_expr, prefix='descendant-or-self::'):
-    if isinstance(css_expr, basestring):
+    if isinstance(css_expr, _basestring):
         match = _el_re.search(css_expr)
         if match is not None:
             return '%s%s' % (prefix, match.group(0).strip())
@@ -513,7 +540,7 @@ def css_to_xpath(css_expr, prefix='descendant-or-self::'):
         "Got None for xpath expression from %s" % repr(css_expr))
     if prefix:
         expr.add_prefix(prefix)
-    return str(expr)
+    return _unicode(expr)
 
 class XPathExpr(object):
 
@@ -528,10 +555,10 @@ class XPathExpr(object):
     def __str__(self):
         path = ''
         if self.prefix is not None:
-            path += str(self.prefix)
+            path += _unicode(self.prefix)
         if self.path is not None:
-            path += str(self.path)
-        path += str(self.element)
+            path += _unicode(self.path)
+        path += _unicode(self.element)
         if self.condition:
             path += '[%s]' % self.condition
         return path
@@ -563,7 +590,7 @@ class XPathExpr(object):
         if self.element == '*':
             # We weren't doing a test anyway
             return
-        self.add_condition("name() = %s" % xpath_repr(self.element))
+        self.add_condition("name() = %s" % xpath_literal(self.element))
         self.element = '*'
 
     def add_star_prefix(self):
@@ -578,7 +605,7 @@ class XPathExpr(object):
         self.star_prefix = True
 
     def join(self, combiner, other):
-        prefix = str(self)
+        prefix = _unicode(self)
         prefix += combiner
         path = (other.prefix or '') + (other.path or '')
         # We don't need a star prefix if we are joining to this other
@@ -604,16 +631,26 @@ class XPathExprOr(XPathExpr):
 
     def __str__(self):
         prefix = self.prefix or ''
-        return ' | '.join([prefix + str(i) for i in self.items])
+        return ' | '.join(["%s%s" % (prefix,i) for i in self.items])
 
-def xpath_repr(s):
-    # FIXME: I don't think this is right, but lacking any reasonable
-    # specification on what XPath literals look like (which doesn't seem
-    # to be in the XPath specification) it is hard to do 'right'
+split_at_single_quotes = re.compile("('+)").split
+
+def xpath_literal(s):
     if isinstance(s, Element):
         # This is probably a symbol that looks like an expression...
         s = s._format_element()
-    return repr(str(s))
+    else:
+        s = _unicode(s)
+    if "'" not in s:
+        s = "'%s'" % s
+    elif '"' not in s:
+        s = '"%s"' % s
+    else:
+        s = "concat(%s)" % ','.join([
+            (("'" in part) and '"%s"' or "'%s'") % part
+            for part in split_at_single_quotes(s) if part
+            ])
+    return s
 
 ##############################
 ## Parsing functions
@@ -623,9 +660,15 @@ def parse(string):
     stream.source = string
     try:
         return parse_selector_group(stream)
-    except SelectorSyntaxError, e:
-        e.args = tuple(["%s at %s -> %s" % (
-            e, stream.used, list(stream))])
+    except SelectorSyntaxError:
+        import sys
+        e = sys.exc_info()[1]
+        message = "%s at %s -> %r" % (
+            e, stream.used, stream.peek())
+        e.msg = message
+        if sys.version_info < (2,6):
+            e.message = message
+        e.args = tuple([message])
         raise
 
 def parse_selector_group(stream):
@@ -650,9 +693,16 @@ def parse_selector(stream):
         elif peek in ('+', '>', '~'):
             # A combinator
             combinator = stream.next()
+            # Ignore optional whitespace after a combinator
+            while stream.peek() == ' ':
+                stream.next()
         else:
             combinator = ' '
+        consumed = len(stream.used)
         next_selector = parse_simple_selector(stream)
+        if consumed == len(stream.used):
+            raise SelectorSyntaxError(
+                "Expected selector, got '%s'" % stream.peek())
         result = CombinedSelector(result, combinator, next_selector)
     return result
 
@@ -664,14 +714,14 @@ def parse_simple_selector(stream):
         next = stream.next()
         if next != '*' and not isinstance(next, Symbol):
             raise SelectorSyntaxError(
-                "Expected symbol, got %r" % next)
+                "Expected symbol, got '%s'" % next)
         if stream.peek() == '|':
             namespace = next
             stream.next()
             element = stream.next()
             if element != '*' and not isinstance(next, Symbol):
                 raise SelectorSyntaxError(
-                    "Expected symbol, got %r" % next)
+                    "Expected symbol, got '%s'" % next)
         else:
             namespace = '*'
             element = next
@@ -698,14 +748,14 @@ def parse_simple_selector(stream):
             next = stream.next()
             if not next == ']':
                 raise SelectorSyntaxError(
-                    "] expected, got %r" % next)
+                    "] expected, got '%s'" % next)
             continue
         elif peek == ':' or peek == '::':
             type = stream.next()
             ident = stream.next()
             if not isinstance(ident, Symbol):
                 raise SelectorSyntaxError(
-                    "Expected symbol, got %r" % ident)
+                    "Expected symbol, got '%s'" % ident)
             if stream.peek() == '(':
                 stream.next()
                 peek = stream.peek()
@@ -719,7 +769,7 @@ def parse_simple_selector(stream):
                 next = stream.next()
                 if not next == ')':
                     raise SelectorSyntaxError(
-                        "Expected ), got %r and %r"
+                        "Expected ')', got '%s' and '%s'"
                         % (next, selector))
                 result = Function(result, type, ident, selector)
             else:
@@ -753,11 +803,11 @@ def parse_attrib(selector, stream):
     op = stream.next()
     if not op in ('^=', '$=', '*=', '=', '~=', '|=', '!='):
         raise SelectorSyntaxError(
-            "Operator expected, got %r" % op)
+            "Operator expected, got '%s'" % op)
     value = stream.next()
     if not isinstance(value, (Symbol, String)):
         raise SelectorSyntaxError(
-            "Expected string or symbol, got %r" % value)
+            "Expected string or symbol, got '%s'" % value)
     return Attrib(selector, namespace, attrib, op, value)
 
 def parse_series(s):
@@ -801,17 +851,17 @@ def parse_series(s):
 ## Tokenizing
 ############################################################
 
-_whitespace_re = re.compile(r'\s+')
+_match_whitespace = re.compile(r'\s+', re.UNICODE).match
 
-_comment_re = re.compile(r'/\*.*?\*/', re.S)
+_replace_comments = re.compile(r'/\*.*?\*/', re.DOTALL).sub
 
-_count_re = re.compile(r'[+-]?\d*n(?:[+-]\d+)?')
+_match_count_number = re.compile(r'[+-]?\d*n(?:[+-]\d+)?').match
 
 def tokenize(s):
     pos = 0
-    s = _comment_re.sub('', s)
+    s = _replace_comments('', s)
     while 1:
-        match = _whitespace_re.match(s, pos=pos)
+        match = _match_whitespace(s, pos=pos)
         if match:
             preceding_whitespace_pos = pos
             pos = match.end()
@@ -819,7 +869,7 @@ def tokenize(s):
             preceding_whitespace_pos = 0
         if pos >= len(s):
             return
-        match = _count_re.match(s, pos=pos)
+        match = _match_count_number(s, pos=pos)
         if match and match.group() != 'n':
             sym = s[pos:match.end()]
             yield Symbol(sym, pos)
@@ -832,7 +882,7 @@ def tokenize(s):
             pos += 2
             continue
         if c in '>+~,.*=[]()|:#':
-            if c in '.#' and preceding_whitespace_pos > 0:
+            if c in '.#[' and preceding_whitespace_pos > 0:
                 yield Token(' ', preceding_whitespace_pos)
             yield Token(c, pos)
             pos += 1
@@ -848,6 +898,28 @@ def tokenize(s):
         yield Symbol(sym, old_pos)
         continue
 
+split_at_string_escapes = re.compile(r'(\\(?:%s))'
+                                     % '|'.join(['[A-Fa-f0-9]{1,6}(?:\r\n|\s)?',
+                                                 '[^A-Fa-f0-9]'])).split
+
+def unescape_string_literal(literal):
+    substrings = []
+    for substring in split_at_string_escapes(literal):
+        if not substring:
+            continue
+        elif '\\' in substring:
+            if substring[0] == '\\' and len(substring) > 1:
+                substring = substring[1:]
+                if substring[0] in '0123456789ABCDEFabcdef':
+                    # int() correctly ignores the potentially trailing whitespace
+                    substring = _unichr(int(substring, 16))
+            else:
+                raise SelectorSyntaxError(
+                    "Invalid escape sequence %r in string %r"
+                    % (substring.split('\\')[1], literal))
+        substrings.append(substring)
+    return ''.join(substrings)
+
 def tokenize_escaped_string(s, pos):
     quote = s[pos]
     assert quote in ('"', "'")
@@ -860,13 +932,13 @@ def tokenize_escaped_string(s, pos):
                 "Expected closing %s for string in: %r"
                 % (quote, s[start:]))
         result = s[start:next]
-        try:
-            result = result.decode('unicode_escape')
-        except UnicodeDecodeError:
-            # Probably a hanging \
+        if result.endswith('\\'):
+            # next quote character is escaped
             pos = next+1
-        else:
-            return result, next+1
+            continue
+        if '\\' in result:
+            result = unescape_string_literal(result)
+        return result, next+1
     
 _illegal_symbol = re.compile(r'[^\w\\-]', re.UNICODE)
 
@@ -886,8 +958,10 @@ def tokenize_symbol(s, pos):
         result = s[start:match.start()]
         pos = match.start()
     try:
-        result = result.decode('unicode_escape')
-    except UnicodeDecodeError, e:
+        result = result.encode('ASCII', 'backslashreplace').decode('unicode_escape')
+    except UnicodeDecodeError:
+        import sys
+        e = sys.exc_info()[1]
         raise SelectorSyntaxError(
             "Bad symbol %r: %s" % (result, e))
     return result, pos
@@ -900,6 +974,11 @@ class TokenStream(object):
         self.source = source
         self.peeked = None
         self._peeking = False
+        try:
+            self.next_token = self.tokens.next
+        except AttributeError:
+            # Python 3
+            self.next_token = self.tokens.__next__
 
     def next(self):
         if self._peeking:
@@ -908,7 +987,7 @@ class TokenStream(object):
             return self.peeked
         else:
             try:
-                next = self.tokens.next()
+                next = self.next_token()
                 self.used.append(next)
                 return next
             except StopIteration:
@@ -920,7 +999,7 @@ class TokenStream(object):
     def peek(self):
         if not self._peeking:
             try:
-                self.peeked = self.tokens.next()
+                self.peeked = self.next_token()
             except StopIteration:
                 return None
             self._peeking = True
